#!/usr/bin/env python3

import os
import sys
import json

import singer
from singer import utils
from singer.catalog import Catalog, CatalogEntry, Schema
from . import streams
from .context import Context
from . import schemas

REQUIRED_CONFIG_KEYS = ["token", "metric"]

LOGGER = singer.get_logger()

#def check_authorization(atx):
#    atx.client.get('/settings')


# with tap-emarsys, they do it this way where the catalog is read in from a call to the api
#  but with the odd frontapp structure, we won't do that here
# we never use atx in here since the schema is from file
#  but we would use it if we pulled schema from the API
# def discover(atx):
def discover():
    catalog = Catalog([])
    for tap_stream_id in schemas.STATIC_SCHEMA_STREAM_IDS:
        #print("tap stream id=",tap_stream_id)
        schema = Schema.from_dict(schemas.load_schema(tap_stream_id))
        metadata = []
        for field_name in schema.properties.keys():
            #print("field name=",field_name)
            if field_name in schemas.PK_FIELDS[tap_stream_id]:
                inclusion = 'automatic'
            else:
                inclusion = 'available'
            metadata.append({
                'metadata': {
                    'inclusion': inclusion
                },
                'breadcrumb': ['properties', field_name]
            })
        catalog.streams.append(CatalogEntry(
            stream=tap_stream_id,
            tap_stream_id=tap_stream_id,
            key_properties=schemas.PK_FIELDS[tap_stream_id],
            schema=schema,
            metadata=metadata
        ))
    return catalog


# this is already defined in schemas.py though w/o dependencies.  do we keep this for the sync?
def load_schema(tap_stream_id):
    path = "schemas/{}.json".format(tap_stream_id)
    schema = utils.load_json(get_abs_path(path))
    dependencies = schema.pop("tap_schema_dependencies", [])
    refs = {}
    for sub_stream_id in dependencies:
        refs[sub_stream_id] = load_schema(sub_stream_id)
    if refs:
        singer.resolve_schema_references(schema, refs)
    return schema


def sync(atx):
    for tap_stream_id in schemas.STATIC_SCHEMA_STREAM_IDS:
        schemas.load_and_write_schema(tap_stream_id)

    streams.sync_selected_streams(atx)


@utils.handle_top_exception(LOGGER)
def main():
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)
    atx = Context(args.config, args.state)
    if args.discover:
        # the schema is static from file so we don't need to pass in atx for connection info.
        catalog = discover()
        json.dump(catalog.to_dict(), sys.stdout)
    else:
        atx.catalog = Catalog.from_dict(args.properties) \
            if args.properties else discover()
        sync(atx)

if __name__ == "__main__":
    main()



import os
import re

import singer
from singer import utils

class IDS(object): # pylint: disable=too-few-public-methods
    TEAM_TABLE = 'team_table'

STATIC_SCHEMA_STREAM_IDS = [
    IDS.TEAM_TABLE
]

PK_FIELDS = {
    IDS.TEAM_TABLE: ['analytics_date', 'analytics_range', 'teammate_v']
}

def normalize_fieldname(fieldname):
    fieldname = fieldname.lower()
    fieldname = re.sub(r'[\s\-]', '_', fieldname)
    return re.sub(r'[^a-z0-9_]', '', fieldname)


# the problem with the schema we see coming from team_table is that it's a little inconsistent:
# {"t":"str","v":"All","p":"All"},{"t":"num","v":306,"p":465},{"t":"num","v":2.65,"p":2.39},...
# ,[{"t":"teammate","v":"Andrew","url":"/api/1/companies/theguild_co/team/andrew","id":253419},...
# so we see that the type = teammate is different when it covers All team members
# also we see the schema where type = num or dur is actually a triplet of type, value, and previous
# so it looks like we need to hardcode those anomalies into this file

def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)

def load_schema(tap_stream_id):
    path = 'schemas/{}.json'.format(tap_stream_id)
    #print("schema path=",path)
    return utils.load_json(get_abs_path(path))

def load_and_write_schema(tap_stream_id):
    schema = load_schema(tap_stream_id)
    singer.write_schema(tap_stream_id, schema, PK_FIELDS[tap_stream_id])



import json
import time

import requests
import backoff
import singer
from singer import metrics

LOGGER = singer.get_logger()

class RateLimitException(Exception):
    pass

class MetricsRateLimitException(Exception):
    pass

class Client(object):
    BASE_URL = 'https://api2.frontapp.com'

    def __init__(self, config):
        self.token = 'Bearer ' + config.get('token')
        self.metric = config.get('metric')
        self.session = requests.Session()

        self.calls_remaining = None
        self.limit_reset = None


    def url(self, path):
        return self.BASE_URL + path

    @backoff.on_exception(backoff.expo,
                          RateLimitException,
                          max_tries=10,
                          factor=2)
    def request(self, method, path, **kwargs):
        if self.calls_remaining is not None and self.calls_remaining == 0:
            wait = self.limit_reset - int(time.monotonic())
            if 0 < wait <= 300:
                time.sleep(wait)

        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        if self.token:
            kwargs['headers']['Authorization'] = self.token

        kwargs['headers']['Content-Type'] = 'application/json'

        if 'endpoint' in kwargs:
            endpoint = kwargs['endpoint']
            del kwargs['endpoint']
            with metrics.http_request_timer(endpoint) as timer:
                response = requests.request(method, self.url(path), **kwargs)
                timer.tags[metrics.Tag.http_status_code] = response.status_code

                # frontapp's API takes an initial request then needs 1 sec to produce the report
                #so here we just run the request again
                time.sleep(2)
                response = requests.request(method, self.url(path), **kwargs)

        else:
            response = requests.request(method, self.url(path), **kwargs)
            time.sleep(2)
            response = requests.request(method, self.url(path), **kwargs)

        #print('final3 url=',response.url)
        self.calls_remaining = int(response.headers['X-Ratelimit-Remaining'])
        self.limit_reset = int(float(response.headers['X-Ratelimit-Reset']))

        if response.status_code in [429, 503]:
            raise RateLimitException()
        if response.status_code == 423:
            raise MetricsRateLimitException()
        try:
            response.raise_for_status()
        except:
            LOGGER.error('{} - {}'.format(response.status_code, response.text))
            raise

        if len(response.json()['metrics']) > 0:
            return response.json()['metrics'][0]['rows']
        else:
            return {}

    def get(self, path, **kwargs):
        return self.request('get', path, **kwargs)

    def post(self, path, data, **kwargs):
        kwargs['data'] = json.dumps(data)
        return self.request('post', path, **kwargs)



import time
import datetime

import pendulum
import singer
from singer import metadata
from singer.bookmarks import write_bookmark, reset_stream
from ratelimit import limits, sleep_and_retry, RateLimitException
from backoff import on_exception, expo, constant

from .schemas import (
    IDS
)
from .http import MetricsRateLimitException

LOGGER = singer.get_logger()

MAX_METRIC_JOB_TIME = 1800
METRIC_JOB_POLL_SLEEP = 1

def count(tap_stream_id, records):
    with singer.metrics.record_counter(tap_stream_id) as counter:
        counter.increment(len(records))

def write_records(tap_stream_id, records):
    singer.write_records(tap_stream_id, records)
    count(tap_stream_id, records)

def get_date_and_integer_fields(stream):
    date_fields = []
    integer_fields = []
    for prop, json_schema in stream.schema.properties.items():
        _type = json_schema.type
        if isinstance(_type, list) and 'integer' in _type or \
            _type == 'integer':
            integer_fields.append(prop)
        elif json_schema.format == 'date-time':
            date_fields.append(prop)
    return date_fields, integer_fields

def base_transform(date_fields, integer_fields, obj):
    new_obj = {}
    for field, value in obj.items():
        if value == '':
            value = None
        elif field in integer_fields and value is not None:
            value = int(value)
        elif field in date_fields and value is not None:
            value = pendulum.parse(value).isoformat()
        new_obj[field] = value
    return new_obj

# not using this for now since we're only initially building this for the team_table
# and want all its data, but will leave in for further enhancement
def select_fields(mdata, obj):
    new_obj = {}
    for key, value in obj.items():
        field_metadata = mdata.get(('properties', key))
        if field_metadata and \
            (field_metadata.get('selected') is True or \
            field_metadata.get('inclusion') == 'automatic'):
            new_obj[key] = value
    return new_obj

@on_exception(constant, MetricsRateLimitException, max_tries=5, interval=60)
@on_exception(expo, RateLimitException, max_tries=5)
@sleep_and_retry
@limits(calls=1, period=61) # 60 seconds needed to be padded by 1 second to work
def get_metric(atx, metric, start_date, end_date):
    LOGGER.info('Metrics query - metric: {} start_date: {} end_date: {} '.format(
        metric,
        start_date,
        end_date))
    return atx.client.get('/analytics', params={'start': start_date, \
            'end': end_date, 'metrics[]':metric}, endpoint='analytics')

def sync_metric(atx, metric, incremental_range, start_date, end_date):
    with singer.metrics.job_timer('daily_aggregated_metric'):
        start = time.monotonic()
        start_date_formatted = datetime.datetime.utcfromtimestamp(start_date).strftime('%Y-%m-%d')
        # we've really moved this functionality to the request in the http script
        #so we don't expect that this will actually have to run mult times
        while True:
            if (time.monotonic() - start) >= MAX_METRIC_JOB_TIME:
                raise Exception('Metric job timeout ({} secs)'.format(
                    MAX_METRIC_JOB_TIME))
            data = get_metric(atx, metric, start_date, end_date)
            if data != '':
                break
            else:
                time.sleep(METRIC_JOB_POLL_SLEEP)

    data_rows = []
    # transform the team_table data
    if metric == 'team_table':
        rnum = 0
        for row in data:
            rnum += 1
            # the first row returned from frontapp is an aggregate row
            # and has a slightly different form
            if rnum == 1:
                data_rows.append({
                    "analytics_date": start_date_formatted,
                    "analytics_range": incremental_range,
                    "teammate_v": row[0]['v'],
                    "teammate_url": "",
                    "teammate_id": 0,
                    "teammate_p": row[0]['p'],
                    "num_conversations_v": row[1]['v'],
                    "num_conversations_p": row[1]['p'],
                    "avg_message_conversations_v": row[2]['v'],
                    "avg_message_conversations_p": row[2]['p'],
                    "avg_reaction_time_v": row[3]['v'],
                    "avg_reaction_time_p": row[3]['p'],
                    "avg_first_reaction_time_v": row[4]['v'],
                    "avg_first_reaction_time_p": row[4]['p'],
                    "num_messages_v": row[5]['v'],
                    "num_messages_p": row[5]['p'],
                    "num_sent_v": row[6]['v'],
                    "num_sent_p": row[6]['p'],
                    "num_replied_v": row[7]['v'],
                    "num_replied_p": row[7]['p'],
                    "num_composed_v": row[8]['v'],
                    "num_composed_p": row[8]['p']
                    })
            else:
                data_rows.append({
                    "analytics_date": start_date_formatted,
                    "analytics_range": incremental_range,
                    "teammate_v": row[0]['v'],
                    "teammate_url": row[0]['url'],
                    "teammate_id": row[0]['id'],
                    "teammate_p": row[0]['v'],
                    "num_conversations_v": row[1]['v'],
                    "num_conversations_p": row[1]['p'],
                    "avg_message_conversations_v": row[2]['v'],
                    "avg_message_conversations_p": row[2]['p'],
                    "avg_reaction_time_v": row[3]['v'],
                    "avg_reaction_time_p": row[3]['p'],
                    "avg_first_reaction_time_v": row[4]['v'],
                    "avg_first_reaction_time_p": row[4]['p'],
                    "num_messages_v": row[5]['v'],
                    "num_messages_p": row[5]['p'],
                    "num_sent_v": row[6]['v'],
                    "num_sent_p": row[6]['p'],
                    "num_replied_v": row[7]['v'],
                    "num_replied_p": row[7]['p'],
                    "num_composed_v": row[8]['v'],
                    "num_composed_p": row[8]['p']
                    })

    write_records(metric, data_rows)

def write_metrics_state(atx, metric, date_to_resume):
    write_bookmark(atx.state, metric, 'date_to_resume', date_to_resume.to_datetime_string())
    atx.write_state()

def sync_metrics(atx, metric):
    incremental_range = atx.config.get('incremental_range')
    stream = atx.catalog.get_stream(metric)
    bookmark = atx.state.get('bookmarks', {}).get(metric, {})

    LOGGER.info('metric: {} '.format(metric))
    mdata = metadata.to_map(stream.metadata)

    # start_date is defaulted in the config file 2018-01-01
    # if there's no default date and it gets set to now, then start_date will have to be
    #   set to the prior business day/hour before we can use it.
    now = datetime.datetime.now()
    if incremental_range == "daily":
        s_d = now.replace(hour=0, minute=0, second=0, microsecond=0)
        start_date = pendulum.parse(atx.config.get('start_date', s_d + datetime.timedelta(days=-1, hours=0)))
    elif incremental_range == "hourly":
        s_d = now.replace(minute=0, second=0, microsecond=0)
        start_date = pendulum.parse(atx.config.get('start_date', s_d + datetime.timedelta(days=0, hours=-1)))
    LOGGER.info('start_date: {} '.format(start_date))


    # end date is not usually specified in the config file by default so end_date is now.
    # if end date is now, we will have to truncate them
    # to the nearest day/hour before we can use it.
    if incremental_range == "daily":
        e_d = now.replace(hour=0, minute=0, second=0, microsecond=0).strftime("%Y-%m-%d %H:%M:%S")
        end_date = pendulum.parse(atx.config.get('end_date', e_d))
    elif incremental_range == "hourly":
        e_d = now.replace(minute=0, second=0, microsecond=0).strftime("%Y-%m-%d %H:%M:%S")
        end_date = pendulum.parse(atx.config.get('end_date', e_d))
    LOGGER.info('end_date: {} '.format(end_date))

    # if the state file has a date_to_resume, we use it as it is.
    # if it doesn't exist, we overwrite by start date
    s_d = start_date.strftime("%Y-%m-%d %H:%M:%S")
    last_date = pendulum.parse(bookmark.get('date_to_resume', s_d))
    LOGGER.info('last_date: {} '.format(last_date))


    # no real reason to assign this other than the naming
    # makes better sense once we go into the loop
    current_date = last_date

    while current_date <= end_date:
        if incremental_range == "daily":
            next_date = current_date + datetime.timedelta(days=1, hours=0)
        elif incremental_range == "hourly":
            next_date = current_date + datetime.timedelta(days=0, hours=1)

        ut_current_date = int(current_date.timestamp())
        LOGGER.info('ut_current_date: {} '.format(ut_current_date))
        ut_next_date = int(next_date.timestamp())
        LOGGER.info('ut_next_date: {} '.format(ut_next_date))
        sync_metric(atx, metric, incremental_range, ut_current_date, ut_next_date)
        # if the prior sync is successful it will write the date_to_resume bookmark
        write_metrics_state(atx, metric, next_date)
        current_date = next_date

    reset_stream(atx.state, metric)

def sync_selected_streams(atx):
    selected_streams = atx.selected_stream_ids

    # last_synced_stream = atx.state.get('last_synced_stream')

    if IDS.TEAM_TABLE in selected_streams:
        sync_metrics(atx, 'team_table')

    # add additional analytics here



from datetime import datetime, date

import singer
from singer import bookmarks as bks_, metadata

from .http import Client

class Context(object):
    """Represents a collection of global objects necessary for performing
    discovery or for running syncs. Notably, it contains

    - config  - The JSON structure from the config.json argument
    - state   - The mutable state dict that is shared among streams
    - client  - An HTTP client object for interacting with the API
    - catalog - A singer.catalog.Catalog. Note this will be None during
                discovery.
    """
    def __init__(self, config, state):
        self.config = config
        self.state = state
        self.client = Client(config)
        self._catalog = None
        self.selected_stream_ids = None
        self.now = datetime.utcnow()

    @property
    def catalog(self):
        return self._catalog

    @catalog.setter
    def catalog(self, catalog):
        self._catalog = catalog
        self.selected_stream_ids = set()
        for stream in catalog.streams:
            mdata = metadata.to_map(stream.metadata)
            root_metadata = mdata.get(())
            if root_metadata and root_metadata.get('selected') is True:
                self.selected_stream_ids.add(stream.tap_stream_id)

    def get_bookmark(self, path):
        return bks_.get_bookmark(self.state, *path)

    def set_bookmark(self, path, val):
        if isinstance(val, date):
            val = val.isoformat()
        bks_.write_bookmark(self.state, path[0], path[1], val)

    def get_offset(self, path):
        off = bks_.get_offset(self.state, path[0])
        return (off or {}).get(path[1])

    def set_offset(self, path, val):
        bks_.set_offset(self.state, path[0], path[1], val)

    def clear_offsets(self, tap_stream_id):
        bks_.clear_offset(self.state, tap_stream_id)

    def write_state(self):
        singer.write_state(self.state)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name="tap-frontapp",
    version="0.3.2",
    description="Singer.io tap for extracting data from the FrontApp API",
    author="bytcode.io",
    url="http://singer.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    install_requires=[
        "singer-python>=5.1.1",
        "pendulum",
        "ratelimit",
        "backoff",
        "requests",
    ],
    entry_points="""
    [console_scripts]
    tap-frontapp=tap_frontapp:main
    """,
    packages=find_packages(),
    package_data = {
        "schemas": ["tap_frontapp/schemas/*.json"]
    },
    include_package_data=True
)
